#!/usr/bin/env ruby

require 'rdparse.rb'
require 'xtatInterpreter.rb'
#require '/usr/lib64/ruby/1.8/apache/rdparse.rb'
#require '/usr/lib64/ruby/1.8/apache/xtatInterpreter.rb'


##############################################################################
#
#  Xtat language
#  Verson: 1.1,2
#  Last update: 2010-05-27
#  Creator: Mikael Andersson och Benjamin Wallin
#
##############################################################################

class Xtat

  def initialize
    @xtatParser = Parser.new("xtat") do

      # <--Start Tokens-->

      # Increment/decrement operators
      token(/\+\+/) {|m| m }
      token(/\-\-/) {|m| m }
      # Logical operators
      token(/\|/) {|m| m }
      token(/\&/) {|m| m }
      # Comp operators
      token(/\!\=/) {|m| m }
      token(/\<\=/) {|m| m }
      token(/\>\=/) {|m| m }
      token(/\</) {|m| m }
      token(/\>/) {|m| m }
      token(/\=\=/) {|m|m }
      # Comment-tags
      token(/\/\*/) {|m| m }
      token(/\*\//) {|m| m } 
      # Assignment_operator
      token(/\=/) {|m| m }
      # String-token
      token(/\'/) {|m| m }
      # Whitespace characters
      token(/\s+/)
      # Alphanumeric characters plus "_" ([A-Za-z0-9_])
      # Also Swedish characters and more
      token(/[\w|||||||\?|\!]+/) {|m| m }
      # Digits
      token(/\d+/) {|m| m.to_i }
      token(/./) {|m| m }
   
      # <--End tokens-->


      # <--Start rules-->

      start :program do
	match(:stmt_list) {|stmt_list| Program.new(stmt_list).eval }
      end
      
      rule :stmt_list do
        # Pie chart.
        match(:stat_circle) {|circle| StmtList_class.new(circle) }
        match(:stmt_list,:stat_circle) {|stmt_list,circle| StmtList_class.new([stmt_list,circle]) }
        # Bar chart.
        match(:stat_bar) {|bar| StmtList_class.new(bar) }
        match(:stmt_list,:stat_bar) {|stmt_list,bar| StmtList_class.new([stmt_list,bar]) }
        # Assignment.
        match(:assignment) {|assignment| StmtList_class.new(assignment) }
        match(:stmt_list, :assignment)  {|stmt_list,assignment| StmtList_class.new([stmt_list,assignment]) }
        # Functions.
        match(:stmt_list,:function) {|stmt_list,function| StmtList_class.new([stmt_list,function]) }
        match(:function) {|function| StmtList_class.new(function) }
        # Function calls.
        match(:stmt_list,:function_call) {|stmt_list,function_call| StmtList_class.new([stmt_list,function_call]) }
        match(:function_call){|function_call| function_call }
	# Iteration functions.
        match(:stmt_list,:list_iter) {|stmt_list,list_iter| StmtList_class.new([stmt_list,list_iter]) }
	match(:list_iter) {|list_iter| StmtList_class.new(list_iter) }
	# If-statements.
        match(:if_stmt) {|if_stmt| StmtList_class.new(if_stmt) }
	match(:stmt_list,:if_stmt) {|stmt_list,if_stmt| StmtList_class.new([stmt_list,if_stmt]) }
        # Comments.
        match(:comment) {|comment| StmtList_class.new(comment) }
        match(:stmt_list,:comment) {|stmt_list,comment| StmtList_class.new([stmt_list,comment]) }
        # Print-statements.
        match(:print_stmt) {|print_list| StmtList_class.new(print_list) }
        match(:stmt_list,:print_stmt)  {|stmt_list,print_list| StmtList_class.new([stmt_list,print_list]) }
        # Variable types.
        match(:var_list) {|var_list| StmtList_class.new(var_list) }
        match(:stmt_list,:var_list) {|stmt_list,var_list| StmtList_class.new([stmt_list,var_list]) }
      end

      # To create a pie chart
      rule :stat_circle do
        match("circle","(",:stat_args,")") {|_,_,stat_args,_| Stat_circle.new(stat_args) }
      end

      # To create a bar chart
      rule :stat_bar do
        match("bar","(",:stat_args,")") {|_,_,stat_args,_| Stat_bar.new(stat_args) }
      end
      
      # Input values for the diagrams
      rule :stat_args do
        match(:var,",",:var,",","st") {|var,_,var2,_,type| Stat_args.new(var,var2,type) }
        match(:var,",",:var,",","%") {|var,_,var2,_,type| Stat_args.new(var,var2,type) }
        match(:var,",",:var) {|var,_,var2| Stat_args.new(var,var2) }
        match(:var) {|var| Stat_args.new(var) }
      end

      rule :function do
        # With stmt_list.
        match('func',:func_name,'(',:args,')',:stmt_list,:return_stmt,"end") {|_,func_name,_,args,_,stmt_list,return_val,_| Func_class.new(func_name,args,stmt_list,return_val) }
        # Without stmt_list.
        match('func',:func_name,'(',:args,')',:return_stmt,"end") {|_,func_name,_,args,return_val,_| Func_class.new(func_name,args,'',return_val) }
      end
      
      rule :return_stmt do
        match('return',:args) {|_,args| args }
      end
      
      rule :function_call do
        match(:var, :assignment_operator, :func_name,'(',:args,')') {|var,_,func_name,_,args,_| FuncCall_class.new(func_name,args,var) }
      end
   
      rule :list_iter do
        match(:it_for) {|it_for| it_for }
        match(:it_while) {|it_while| it_while }
      end
      
      rule :it_for do
        # With assignment and increment.
        match('for','(',:assignment,',',:expression,',',:increment,')',:stmt_list,"end") {|_,_,var,_,expr,_,count,_,stmt_list,_| For_class.new(var,expr,count,stmt_list) }
        # With assignment and statement list.
        match('for','(',:assignment,',',:expression,',',:stmt_list,')',:stmt_list,"end") {|_,_,var,_,expr,_,count,_,stmt_list,_| For_class.new(var,expr,count,stmt_list) }
        # With variable types and statement list.
        match('for','(',:var_list,',',:expression,',',:stmt_list,')',:stmt_list,"end") {|_,_,var,_,expr,_,count,_,stmt_list,_| For_class.new(var,expr,count,stmt_list) }
        # With variable types and increment.
        match('for','(',:var_list,',',:expression,',',:increment,')',:stmt_list,"end") {|_,_,var,_,expr,_,count,_,stmt_list,_| For_class.new(var,expr,count,stmt_list) }
      end

      rule :it_while do
        match('while','(',:expression,')',:stmt_list,"end") {|_,_,expression,_,stmt_list,_| While_class.new(expression,stmt_list) }
      end
      
      rule :func_name do
        match(:identifier) {|identifier| identifier }
      end
      
      rule :comment do
        match("/*",:comment_lines,"*/") {|_,comment,_| Comment_class.new(comment) }
      end
      
      #TODO: Fix the white-space-ending problem.
      rule :comment_lines do
        match(:identifier) {|comment| comment }
        match(:comment_lines,:identifier) {|comment,comment2| [comment,comment2] }
      end

      rule :print_stmt do
        # Strings.
        match("print",:string_stmt) {|_,string_stmt| Print_class.new(string_stmt) }
        # Variables.
        match("print",:var_list) {|_,var_list| Print_class.new(var_list) }
      end

      rule :assignment do
        # Increment/decrement vars.
        match(:increment)
        # Matches Arrays.
        match(:var,:assignment_operator,:array) {|name,_,value| Array_class.new(name,value) } 
        match(:const_var,:assignment_operator,:array) {|name,_,value| Array_class.new(name,value) } 
        # Matches bools.
        match(:var,:assignment_operator,:bool) {|name,_,value| Var_class.new(name,value) }
        match(:const_var,:assignment_operator,:bool) {|name,_,value| Var_class.new(name,value) }
        # Matches numbers.
        match(:var,:assignment_operator,:math) {|name,_,value| Var_class.new(name,value) }
        match(:const_var,:assignment_operator,:math) {|name,_,value| Var_class.new(name,value) }
        # Matches strings.
        match(:var,:assignment_operator,:string_stmt) {|name,_,value| Var_class.new(name,value) }
        match(:const_var,:assignment_operator,:string_stmt) {|name,_,value| Var_class.new(name,value) }
      end

      rule :increment do
        # Increment.
        match(:var,"++") {|var_name,_| Increment_class.new(var_name,"increment")}
        # Decrement.
        match(:var,"--") {|var_name,_| Increment_class.new(var_name,"decrement")}
      end

      rule :var_list do
        # Constants.
        match(:const_var) {|const_var| const_var }
        # Normal variables.
        match(:var) {|var| var }
      end

      rule :array do
        match('array','(',:args,')') {|_,_,args,_| args }
        match('array','(',')') 
      end

      rule :var do
        match(:var_sign,:lowercase_var) {|_,lower| Lower_letter.new(lower) }
      end
      
      rule :const_var do
        match(:var_sign,:uppercase_var) {|_,upper| Upper_letter.new(upper) }
      end

      rule :if_stmt do
        # Oneliner if-statement.
        match(:oneliner_if_stmt) {|stmt_list| stmt_list }
        # If-statement.
        match('if','(',:if_expression,')',:stmt_list,"end") {|_,_,expression,_,stmt_list,_| If_class.new(expression,stmt_list) } 
        # Else-statement.
        match('if','(',:if_expression,')',:stmt_list,:else_stmt) {|_,_,expression,_,stmt_list,else_stmt_list| If_class.new(expression,stmt_list,else_stmt_list) }
        # Else-if-statement.
        match('if','(',:if_expression,')',:stmt_list,:else_if_stmt,:else_stmt) {|_,_,expression,_,stmt_list,else_if_stmt_list,else_stmt_list| If_class.new(expression,stmt_list,else_stmt_list,else_if_stmt_list) }
      end
      
      rule :else_if_stmt do
        # One else-if.
        match("elseif","(",:if_expression,")",:stmt_list) {|_,_,expression,_,stmt_list|[expression,stmt_list] }
        # More else-if functions.
        match(:else_if_stmt,"elseif","(",:if_expression,")",:stmt_list) {|exp2,_,_,expression,_,stmt_list|[exp2[0],exp2[1],expression,stmt_list] }
      end

      rule :else_stmt do        
        match("else",:stmt_list,"end") {|_,stmt_list,_| stmt_list }
      end
            
      rule :oneliner_if_stmt do
        match("(",:if_expression,")","?",:stmt_list,":",:stmt_list) {|_,expression,_,_,stmt_list_true,_,stmt_list_false| If_class.new(expression,stmt_list_true,stmt_list_false) }
      end
      
      rule :if_expression do
        match(:expression){|expression|expression}
        match("(",:if_expression,")"){|_,expression,_|expression}
        match(:if_expression,:log_operator,"(",:expression,")"){|expression,log_op,_,expression2,_|[expression,log_op,expression2]}
      end

      rule :expression do 
        match(:var_list, :operator, :var_list) {|expr,op,expr2| Expression_class.new(expr,op,expr2) }
        match(:var_list, :operator, :bool) {|expr,op,expr2| Expression_class.new(expr,op,expr2) }
        match(:var_list, :operator, :digit) {|expr,op,expr2| Expression_class.new(expr,op,expr2) }
        match(:bool, :operator, :var_list) {|expr,op,expr2| Expression_class.new(expr,op,expr2) }
        match(:bool, :operator, :bool) {|expr,op,expr2| Expression_class.new(expr,op,expr2) }
        match(:var_list) {|var| Expression_class.new(var) }
        match(:bool) {|bool| Expression_class.new(bool) }
      end

     rule :math do 
        match(:math,:math_operator,:var_list){|expr, op, expr2| Math_class.new(expr,op,expr2) }
        match(:var_list,:math_operator,:math){|expr, op, expr2| Math_class.new(expr,op,expr2) }
        match(:math,:math_operator,:math){|expr, op, expr2| Math_class.new(expr,op,expr2) }
        match(:math_prio) {|math_prio| math_prio }
      end
      
      rule :math_prio do 
        match(:math_prio,:math_prio_op,:math_prio) {|expr, op, expr2| Math_class.new(expr,op,expr2) }
        match(:math_pare_prio) {|math_pare_prio| math_pare_prio }
      end

      # Numerical priority with parentheses.
      rule :math_pare_prio do
        match("(",:math,")") {|_, expr, _| expr }
        match(:digit) {|digit| digit }
      end

      rule :args do 
        match(:var_list,",",:args) {|var_list,_,args| Args_class.new([var_list,args]) }
        match(:identifier,",",:args) {|identifier,_,args| Args_class.new([identifier,args]) }
        match(:string_stmt,",",:args) {|string_stmt,_,args| Args_class.new([string_stmt,args]) }
        match(:var_list) {|var_list| Args_class.new(var_list) }
        match(:digit)  {|digit| Args_class.new(digit) }
        match(:bool) {|bool| Args_class.new(bool) }
        match(:string_stmt) {|string_stmt| Args_class.new(string_stmt) }
      end
      
      rule :string_stmt do
        match("'",:str_identifier,"'") {|_,identifier,_| StringStmt_class.new(identifier) }
      end

      rule :str_identifier do
        match(:str_identifier, :str_chars) {|identifier,letter| [identifier,letter] }
	match(:str_chars) {|letter| letter }
      end

      rule :str_chars do
        match(:str_charlist) {|lowercase| Lower_letter.new(lowercase) }
      end

      rule :str_charlist do
        match(/[A-Z]/)
        match(/[a-z]/)
        match(/[\<\>\.\,\!\/\\]/)
        match(/[\#\\%\&\(\)\=\?]/)
        match(/[\\`\+\-\;\:\_]/)
        match(/[\@\\$\{\[\]\}]/)
        match(/[\^\*\~\\]/)
        match(/[\d]/)
      end

      rule :identifier do
        match(:identifier, :letter) {|identifier,letter| [identifier,letter] }
        match(:identifier, :digit) {|identifier,digit|  [identifier,digit] }
	match(:letter) {|letter| letter }
	match(:digit) {|digit| digit }
      end

      rule :letter do
        match(:lowercase) {|lowercase| Lower_letter.new(lowercase) }
        match(:uppercase) {|uppercase| Upper_letter.new(uppercase) }
        match(/[_]/)
      end

      rule :lowercase do
        match(/[a-z]/)
      end

      rule :uppercase do
        match(/[A-Z]/)
      end

      rule :lowercase_var do
        match(/[a-z0-9_]/)
      end

      rule :uppercase_var do
        match(/[A-Z0-9_]/)
      end

      rule :var_sign do
        match(/^[\247]/)
      end
      
      rule :operator do
        match(:log_operator) {|op| op }
        match(:comp_operator) {|op| op }
        match(:math_operator) {|op| op }
        match(:assignment_operator) {|op| op }
      end

      rule :log_operator do
        match("|") {|op| Operator.new(op) }
        match("&") {|op| Operator.new(op) }
      end
      
      rule :comp_operator do
        match("==") {|op| Operator.new(op) }
        match("!=") {|op| Operator.new(op) }
        match(">=") {|op| Operator.new(op) }
        match("<=") {|op| Operator.new(op) }
        match("<") {|op| Operator.new(op) }
        match(">") {|op| Operator.new(op) }
      end
 
      rule :math_operator do
        match("**") {|op| Operator.new(op) }
        match("+") {|op| Operator.new(op) } 
        match("-") {|op| Operator.new(op) }
        match("*") {|op| Operator.new(op) }
        match("/") {|op| Operator.new(op) }
      end

      # Priority operators.
      rule :math_prio_op do
        match("*") {|op| Operator.new(op) }
        match("/") {|op| Operator.new(op) }
        match("**") {|op| Operator.new(op) }
      end

      rule :assignment_operator do
        match("=") {|op| Operator.new(op) }
      end

      rule :digit do
        match(/\d/) {|value| Digit.new(value) }
      end

      rule :bool do
        match("true") {|bool| Boolean.new(bool) }
        match("false") {|bool| Boolean.new(bool) }
        match("TRUE") {|bool| Boolean.new(bool) }
        match("FALSE") {|bool| Boolean.new(bool) }
      end

      #<--End rules-->
    end
  end
 
  # Exit-checker (roll-function).
  def done(str)
    ["quit","exit",""].include?(str.chomp)
  end
  
  # To test the language.
  def roll
    print "[Xtat] "
    str = gets
    if done(str) then
      puts "Bye."
    else
      puts "=> #{@xtatParser.parse str}"
      roll
    end
  end

  # Run Xt-script.
  def run(filename)
    @xtatParser.logger.level = Logger::WARN
    @xtatParser.parse filename
    return ""
  end

  # Test whole file.
  def testfile
    @xtatParser.logger.level = Logger::WARN
    testfile = IO.read("testfile.xt")
    @xtatParser.parse testfile
    return ""
  end
end
