#!/usr/bin/env ruby


##############################################################################
#
#  Xtat language Interpreter
#  Verson: 1.1.2
#  Last update: 2010-05-27
#  Creator: Mikael Andersson och Benjamin Wallin
#
##############################################################################

#Global list with all arrays.
$arraylist = Hash.new
#Global list with all variables.
$varlist = Hash.new

# Run Xtat language.
class Program
  def initialize(stmt_list)
    @program_code = stmt_list.eval
  end
  def eval
    # Evaluating program.
    return instance_eval(@program_code) 
  end
end

# Array-class.
class Array_class
  def initialize(var,args='')  
    @var = var.eval.to_s
    @args = args.eval
    # Stores the values as an array in a list with the array-name (var-name) as key.
    $arraylist[@var] = @args.split(',')
    # If the array is empty.
    if (@args == "array")
      @array_code = @var+" = Array.new"
    else
      @array_code = @var+"= ["+@args.to_s+"]"
    end
  end
  def eval
    @array_code
  end
end

# Variable class.
class Var_class
  def initialize(var,value)
    @var = var.eval.to_s
    @value = value
    # Stores the var-value in a list with the var-name as key.
    $varlist[@var] = @value
    @var_code = @var+"="+@value.eval.to_s
  end
  # Returns the var-name.
  def varname
    @var
  end
  def eval
    @var_code
  end
end

# Print class.
class Print_class
  def initialize(string)
    if(string.class == StringStmt_class)
      @string = string.eval.to_s
    else
      # Evaluating the string-class and stores it as a normal Ruby-string.
      if ($varlist[string.eval.to_s] != nil)
        @string = string.eval
      else
        @string = string.eval.to_s
      end
    end
    @print_code = "puts "+@string
  end
  def eval
    @print_code
  end
end

# Increment/Decrement-class
class Increment_class
  def initialize(var_name,type)
    @type = type
    @var_string = var_name.eval.to_s+" = "+var_name.eval.to_s
    if (type == "increment")
      @value = @var_string+" + 1"
    end
    if (type == "decrement")
      @value = @var_string+" - 1"
    end
  end
  def type
    @type
  end
  def eval
    @value
  end
end

# Statements-class.
class StmtList_class
  def initialize(stmtlist)
    if(stmtlist.class == Array)
      stmtlist.flatten!
      @stmt_code = ""
      # Merges the stmt-classes.
      stmtlist.each {|stmt| @stmt_code += stmt.eval.to_s+"\n" }
    else
      @stmt_code = stmtlist.eval.to_s
    end
  end
  def eval
    @stmt_code
  end
end

# Statistics-pie-chart-class.
# Creates a pie chart using the Google chart-API.
class Stat_circle
  # The argument must be an array with numbers.
  def initialize(args)
    @var =  $arraylist[args.var.eval.to_s]
    # Size-formatting.
    @size = "&chs="+(args.size.eval+200).to_s+"x"+args.size.eval.to_s
    # Counts the final value.
    @total = 0
    @var.each do |num|
      @total = @total + num.to_i
    end
    # The text-info beside the circle.
    @panel = ""
    @parts = ""
    @var.each do |num|
      # Parts in percent.
      if(args.type == "%")
        @part_number = ((num.to_f/@total)*100).to_i
      else
        # If the entity "st" is used, the numerical-value will be shown.
        @part_number = num.to_i
      end

      @parts =  @parts + @part_number.to_s+","
      @panel = @panel + num.to_s+"("+@part_number.to_s+args.type+")"+"|"
    end
    # Removes the last characters.
    @parts = @parts[0..-2]
    @panel = @panel[0..-2]

    @dist_graph = "&chd=t:"+@parts
    # Part-names.
    if(args.names == nil)
      @part_names = ""
    else
      @part_names = "&chl="
      # Gets the names in an array which is stored in a variable.
      @names = $arraylist[args.names.eval.to_s]
      @names.each do |name|
        # Sanitizes the string.
        @part_names = @part_names + name.to_s.delete("'") + "|"
      end
      # Removes the last characters.
      @part_names = @part_names[0..-2]
    end
    @chart_info = "&chdl="+@panel
    # Correctly formaterad Gchars-API-URL.
    @charts_url = "http://chart.apis.google.com/chart?cht=p"+@size+@dist_graph+@part_names+@chart_info
    
    @circle_code = "puts \"<br/><img src='"+@charts_url.to_s+"' alt='cirkeldiagram' /><br/>\""
  end
  def eval
    @circle_code
  end
end

# Statistics-bar-chart-class.
# Creates a bar chart using the Google chart-API.
class Stat_bar
  # The argument must be an array with numbers.
  def initialize(args)
    @var =  $arraylist[args.var.eval.to_s]
    # Size-formatting.
    @size = "&chs="+(args.size.eval+100).to_s+"x"+args.size.eval.to_s
    # Counts the final value.
    @total = 0
    @var.each do |num|
      # If the entity "st" is used, the numerical-value will be shown.
      @total = @total + num.to_i
    end
    # The text-info beside the circle.
    @panel = ""
    @parts = ""
    @var.each do |num|
      if(args.type == "%")
        @part_number = ((num.to_f/@total)*100).to_i
      else
        @part_number = num.to_i
      end

      @parts =  @parts + @part_number.to_s+","
      @panel = @panel + num.to_s+"("+@part_number.to_s+"%)"+"|"
    end
    # Removes the last characters.
    @parts = @parts[0..-2]
    @panel = @panel[0..-2]

    @dist_graph = "&chd=t:"+@parts
    # Part-names.
    if(args.names == nil)
      @part_names = "&chxt=x,y"
    else
      @part_names = "&chxt=x,y&chxl=0:"
      # Gets the names in an array which is stored in a variable.
      @names = $arraylist[args.names.eval.to_s]
      @names.each do |name|
        # Sanitizes the string.
        @part_names = @part_names + name.to_s.delete("'") + "|"
      end
      # Removes the last characters.
      @part_names = @part_names[0..-2]
    end
    @chart_info = "&chdl="+@panel
    # Correctly formaterad Gchars-API-URL.    
    @charts_url = "http://chart.apis.google.com/chart?cht=bvs"+@size+@dist_graph+@part_names+@chart_info
    
    @bar_code = "puts \"<br/><img src='"+@charts_url.to_s+"' alt='cirkeldiagram' /><br/>\""
  end
  def eval
    @bar_code
  end
end

# Statistics arguments-class.
class Stat_args
  def initialize(var,names=nil,type="%",size=Digit.new(200))
    @var = var
    @names = names
    @type = type
    @size = size
  end
  def var
    @var
  end
  def names
    @names
  end
  def type
    @type
  end
  def size
    @size
  end
  def eval
  end
end

# Function-class.
class Func_class
  def initialize(name,args,stmt_list='',ret_stat='')
    if(stmt_list == '')
      @func_code = "def "+name.eval.to_s+"("+args.eval.to_s+")\n"+"return "+ret_stat.eval.to_s+"\nend"
    else
      @func_code = "def "+name.eval.to_s+"("+args.eval.to_s+")\n"+stmt_list.eval.to_s+"\n"+"return "+ret_stat.eval.to_s+"\nend"
    end
  end
  def eval
    @func_code
  end
end

# Function call-class.
class FuncCall_class
  def initialize(func_name,args,var='')
    @name = func_name.eval.to_s
    @args = args.eval.to_s
    if(var == '')
      @funccall_code = @name+"("+@args+")"
    else
      @var = var.eval.to_s
      @value = Value_class.new(@name+"("+@args+")")
      # Keep it in a varable, an save it into varlist.
      $varlist[@var] = @value
      @funccall_code = @var+" = "+@value.eval.to_s
    end
  end
  def eval
    @funccall_code
  end
end

# For-loop-class.
class For_class
  def initialize(var,expression,count,stmt_list)
    @for_code = var.eval.to_s+"\nwhile("+expression.eval.to_s+")\n"+stmt_list.eval.to_s+"\n"+count.eval.to_s+"\nend"
  end
  def eval
    @for_code
  end
end

# While-loop-class.
class While_class
  def initialize(expression,stmt_list)
    @while_code = "while("+expression.eval.to_s+")\n"+stmt_list.eval.to_s+"\nend"
  end
  def eval
    @while_code
  end
end

#For if-, else-, and else-if-statments.
class If_class
  def initialize(expression,stmt_list,else_stmt_list='',else_if_stmt_list='')
    if_statements = ""
    else_statements = ""
    
    # If-statement.
   if (stmt_list.class == Array)
      # Returns an one-dimensional array.
      stmt_list.flatten!
      # Separating if-stmt-lines by newline.
      stmt_list.each do |line|
        if_statements += line.eval.chomp+"\n"
      end
    else
      # If falsestmt is a String (e.g. oneliner-if-stmt)
      if_statements = stmt_list.eval.to_s
    end

    # Else-statement.
    if (else_stmt_list.class == Array)
      # Returns an one-dimensional array.
      else_stmt_list.flatten!
      # Separating if-stmt-lines by newline.
      else_stmt_list.each do |line|
        else_statements += line.eval.chomp
      end        
    else
      # If falsestmt is a String (e.g. oneliner-if-stmt)
      else_statements = else_stmt_list.eval.to_s
    end
    
    # Else-if-statement.
    x = 0
    elseif_data = ""
    while(x < else_if_stmt_list.size)
      elseif_statements = ""
      if (else_if_stmt_list[x] != nil)
        if (else_if_stmt_list[x+1].class == Array)
          # Separating stmt-lines by newline.
          else_if_stmt_list[x+1].flatten!
          else_if_stmt_list[x+1].each do |line|
            elseif_statements += line.eval.chomp
          end
        else
          elseif_statements = else_if_stmt_list[x+1].eval.to_s
        end
        # Preparing elseif_var-string for evaluation.
        elseif_data += "elsif("+else_if_stmt_list[x].eval.to_s+")\n"
        elseif_data += elseif_statements.to_s
        x+=2
      end
    end 

    # Preparing if-expression
    if_expression = ""
    if (expression.class == Array)
      expression.each do |line|
        if_expression += line.eval.to_s+" "
      end
    else
      if_expression = expression.eval.to_s
    end

    # Eval-var.
    @if_code = "if ("+if_expression+")\n"+if_statements.to_s+elseif_data.to_s+"\nelse\n"+else_statements.to_s+"\nend"
  end
  def eval
    @if_code
  end
end

# Expression-class. (e.g. §i <= 6)
class Expression_class
  def initialize(expr,operator="",expr2=false)
    if(operator == "")
      @expr_bool = expr.eval.to_s
    else
      @expr_bool = expr.eval.to_s+operator.eval+expr2.eval.to_s
    end
  end
  def eval
    @expr_bool
  end
end

# Math-class - for calculating simple math.
class Math_class
  def initialize(expr,operator,expr2)
    @expr = expr.eval
    @expr2 = expr2.eval
    @operator = operator.eval
  end
  def eval
    case @operator
      when "+" then @expr.to_s+"+"+@expr2.to_s
      when "-" then @expr.to_s+"-"+@expr2.to_s
      when "*" then @expr.to_s+"*"+@expr2.to_s
      when "/" then @expr.to_s+"/"+@expr2.to_s
      # Power of.
      when "**" then @expr.to_s"**"+@expr2.to_s
      else @expr
    end
  end
end

# Argument-class.
class Args_class
  def initialize(args)
    if(args.class == Array)
      # Returns an one-dimensional array.
      args.flatten!
      @args_code = ""
      args.each {|arg| @args_code += arg.eval.to_s+"," }
      # Removes the last comma-char.
      @args_code = @args_code[0..-2]
    else
      @args_code = args.eval.to_s
    end
  end
  def eval
    @args_code
  end
end

# String-statement-class.
class StringStmt_class
  def initialize(string)
    if(string.class == Array)
      # Returns an one-dimensional array.
      string.flatten!
      @string_code = "'"
      string.each {|str| @string_code += str.eval.to_s+" " }
      @string_code += "'"
    else
      @string_code = "'"+string.eval.to_s+"'"
    end
  end
  def eval
    @string_code
  end
end

# Comments-class.
class Comment_class
  def initialize(string)
    if(string.class == Array)
      # Returns an one-dimensional array.
      string.flatten!
      @comment_code = ""
      string.each {|str| @comment_code += "# "+str.eval.to_s+"\n" }
    else
      @comment_code = "# "+string.eval.to_s+"\n"
    end
  end
  def eval
    @comment_code
  end
end

# Upper-letter-class - capitalized letters.
class Upper_letter
  def initialize(letter)
    @upper_letter = letter.upcase
  end
  def eval
    @upper_letter
  end
end

# Lower-letter-class - lower case letters.
class Lower_letter
  def initialize(letter)
    @lower_letter = letter.downcase
  end
  def eval
    @lower_letter
  end
end

# Operator-class - (Math, comp and log).
class Operator
  def initialize(operator)
    @operator = operator.to_s
  end
  def eval
    @operator
  end
end

# Digit-class.
class Digit
  def initialize(value)
    #Converting to integer
    @digits = value.to_i
  end
  def eval
    @digits
  end
end

# Boolean-class.
class Boolean
  def initialize(bool)
    @bool = bool
  end
  def eval
    @bool
  end
end

# Value-class -  converts a value to string.
class Value_class
  def initialize(value)
    @value_code = value.to_s
  end
  def eval
    @value_code
  end
end
